--
-- Handles initialisation, versioning and migration of saved variables
--
-- provides the global functions ArchiveCreate and ArchiveInit
--
-- User: Swift
-- Date: 15.06.13
-- Time: 09:50
-- To change this template use File | Settings | File Templates.
--

local revision = tonumber(string.sub("$Revision: 22 $", 12, -3))
local rss = _G.RShamanStat
if rss.revision < revision then rss.revision = revision end


-- Archive version and code version are separate, as not every code change changes the archive
local archiveVersion = 19



-- Instructions for changing the archive:
-- 1. Increase the archiveVersion
-- 2. Write a function to migrate data from the previous version to the next
-- 3. Add the function with its version number to the 2-dimensional migration table
-- 4. Update the initial archive initialisation routine
-- 5. If data cannot be migrated, delete all migration functions and add ArchiveClear with the new Version
--    to the migration table

-- Initialize the archive
-- returns either a new or a modified version of the passed archive
function rss.LoadArchive(archive)
    if archive == nil then
        archive = rss.CreateArchive(archive)
    end

    print(("RSS: Loading Archive v%d"):format(archive.version))

    if archive.version == nil then
        archive.version = 0
    end

    if archive.version < archiveVersion then
        -- prevent data loss by testing the code prior to running the real migration
        -- TODO: save the old archive under archive.old and provide a way to revert the data
        local proceed = TestMigration();
        if not proceed then
            print("RSS: Migrations are currently not working. Aborting..")
            return archive
        end;

        print(("RSS: Migrating Archived Records from v%d to v%d"):format(archive.version, archiveVersion))
        MigrateArchive(archive)
    end
    return archive
end

-- returns a new empty archive
function rss.CreateArchive()
    archive = { };
    archive.version = archiveVersion;

    return archive
end


function rss.CreateRecord()
        return {
            startTime               =                        0,
            effectiveTotal          =                       .0,
            totalTime               =                       .0,
            criticalTotal           =                       .0,
            masteryTotal            =                       .0,
            incrementalCritical     =                       .0,
            incrementalCriticalAll  =                       .0,
            incrementalMastery      =                       .0,
            weightedTotal           =                       .0,
            isBoss                  =                    false,
            encounterName           =                "Unknown"
        };
end

function rss.CreateAggregateRecord(name, archive, filter)
    local a = rss.CreateRecord()
    a.encounterName = name
    if (#archive > 0) then
        for i, r in ipairs(archive) do
            if (filter(r)) then
                if a.startTime <= 0 or a.startTime > r.startTime then
                    a.startTime = r.startTime
                end

                a.effectiveTotal = a.effectiveTotal + r.effectiveTotal
                a.totalTime = a.totalTime + r.totalTime
                a.criticalTotal = a.criticalTotal + r.criticalTotal
                a.masteryTotal = a.masteryTotal + r.masteryTotal
                a.incrementalCritical = a.incrementalCritical + r.incrementalCritical
                a.incrementalCriticalAll = a.incrementalCriticalAll + r.incrementalCriticalAll
                a.incrementalMastery = a.incrementalMastery + r.incrementalMastery
                a.isBoss = r.isBoss
                a.weightedTotal = a.weightedTotal + r.weightedTotal
            end
        end
    end
    return a
end

function rss.SaveRecord(archive, record)

    table.insert(archive, record)
end


function rss.PrintRecord(record)
    print(("---- %s ----"):format(record.encounterName))
    local healing = record.effectiveTotal / 1000.
    if record.totalTime > 0 then
        local hps = healing / record.totalTime
        print(("RSS: Effective Healing Done: %.1fk ; Effective HPS: %.1fk"):format(healing,hps))
    else
        print(("RSS: Effective Healing Done: %.1fk"):format(healing));
    end
    if record.effectiveTotal > 0 then
        local critPct = 100 * record.criticalTotal / record.effectiveTotal
        local mstPct = 100 * record.masteryTotal / record.effectiveTotal
        -- critInc and critIncAll will eventually be removed
        local critInc = 0.01 / (record.incrementalCritical / record.effectiveTotal)
        local critIncAll = 0.01 / (record.incrementalCriticalAll / record.effectiveTotal)
        local critCombined = 0.01 * (record.criticalTotal * record.incrementalCritical + (1-record.criticalTotal)*record.incrementalCriticalAll) / record.effectiveTotal
        local mstInc = 0.01 / (record.incrementalMastery / record.effectiveTotal)

        print(("RSS: Throughput from Critical: %.2f%%"):format(critPct));
        print(("RSS: Throughput from Mastery:  %.2f%%"):format(mstPct));
        if record.incrementalCritical > 0 then
            print(("RSS: Critical Rating for +1%% Throughput: %.1f"):format(critInc));
        end
        if record.incrementalCriticalAll > 0 then
            print(("RSS: Critical Rating for +1%% Throughput: %.1f"):format(critIncAll));
        end
        -- weighted average of the two methods, hopefully smooths out fluctuations in actual crit rate
        if record.incrementalCritical > 0 and record.incrementalCriticalAll > 0 then
            print(("RSS: Critical Rating for +1%% Throughput: %.1f"):format(critCombined));
        end
        if record.incrementalMastery > 0 then
            print(("RSS: Mastery Rating for +1%% Throughput: %.1f"):format(mstInc));
        end
    end
end



-- Migration Helper function that adds unset keys to a certain default value
-- keys should be formatted as {"newKey1" = .0, "newKey2" = "DefVal", .. }
function AddKeys(archive, keys)
    for i, record in ipairs(archive) do
        for key, value in pairs(keys) do
            if (not archive[i][key]) then
                archive[i][key] = value
            end
        end
    end
end

-- Migration Helper function that will rename a key
-- keys should be formatted as {"newKey1 = "oldKey1, "newKey2" = "oldKey2", .. }
function RenameKeys(archive, keys)
    for i, record in ipairs(archive) do
        for new, old in pairs(keys) do
            archive[i][new] = record[old]
            archive[i][old] = nil
        end
    end
end

-- Migration helper function that passes every record to a function
-- the passed function needs to take a 'record' as it's single parameter
function Map(archive, fn)
    for i, record in ipairs(archive) do
        fn(archive[i])
    end
end

function ArchiveMigration_18(archive)
    RenameKeys(archive, {
        effectiveTotal = "tot",
        totalTime = "timeTotal",
        criticalTotal = "crtTotal",
        masteryTotal = "mstTotal",
        incrementalCritical = "crtInc",
        incrementalCriticalAll = "crtIncAll",
        incrementalMastery = "mstInc",
        isBoss = "boss",
        encounterName = "label",
    })

    AddKeys(archive, { startTime = 0 })
end

function ArchiveMigration_19(archive)
    -- We may be in elemental spec during data migration, but it's the best we can do.
    AddKeys(archive, {
        resurgence = 0.0,
        spellpower = GetSpellBonusHealing(),
        haste = UnitSpellHaste("player"),
        crit = GetSpellCritChance(4),
        mastery = GetMastery()*3,
    })
    -- Cant use AddKeys because we need to create a new instance of {} for every record
    Map(archive, function(record) record.spells = {}; end)
end

-- Don't use keyword indices here to ensure a fixed order when we iterate over it.
local migrations = {
    {v=18, f=ArchiveMigration_18},
    {v=19, f=ArchiveMigration_19},
}

-- modifies the passed archive to be compatible with the current version
function MigrateArchive(archive)

    for i, m in ipairs(migrations) do
        if archive.version < m.v then
            m.f(archive)
            -- update the version after every step, in case the next migration fails
            archive.version = m.v
        end
    end
end


-- A really simple test to see if the migrations are working as intended
function TestMigration()

    testArchive = {
        {
            ["incrementalCritical"] = 0,
            ["totalTime"] = 5.350000143051148,
            ["masteryTotal"] = 696.9884389214185,
            ["incrementalMastery"] = 0.05130951543812761,
            ["effectiveTotal"] = 42980,
            ["encounterName"] = "Ancient Mogu Spirit",
            ["incrementalCriticalAll"] = -0.017,
            ["weightedTotal"] = 1129.077490359744,
            ["startTime"] = 0,
            ["isBoss"] = false,
            ["criticalTotal"] = 0,
        }, -- [1]
        {
            ["incrementalCritical"] = 2.129230315292802,
            ["totalTime"] = 29.36299991607666,
            ["masteryTotal"] = 12436.08042957493,
            ["incrementalMastery"] = 0.9154947559510546,
            ["effectiveTotal"] = 160238,
            ["encounterName"] = "Ancient Mogu Spirit",
            ["incrementalCriticalAll"] = 2.15583,
            ["weightedTotal"] = 19941.11056392469,
            ["startTime"] = 0,
            ["isBoss"] = false,
            ["criticalTotal"] = 22705,
        }, -- [2]
        {
            ["incrementalCritical"] = 1.028886222069833,
            ["totalTime"] = 23.74300003051758,
            ["masteryTotal"] = 6294.827872376129,
            ["incrementalMastery"] = 0.4634001797760852,
            ["effectiveTotal"] = 137943,
            ["encounterName"] = "Ancient Mogu Spirit",
            ["incrementalCriticalAll"] = 1.463983333333334,
            ["weightedTotal"] = 9982.794133054404,
            ["startTime"] = 0,
            ["isBoss"] = false,
            ["criticalTotal"] = 10549.5,
        }, -- [3]
        {
            ["incrementalCritical"] = 0,
            ["totalTime"] = 18.40700006484985,
            ["masteryTotal"] = 734.9166686378831,
            ["incrementalMastery"] = 0.05410164078699174,
            ["effectiveTotal"] = 34683,
            ["encounterName"] = "Ancient Mogu Spirit",
            ["incrementalCriticalAll"] = 0.02463333333333333,
            ["weightedTotal"] = 1121.215967520285,
            ["startTime"] = 0,
            ["isBoss"] = false,
            ["criticalTotal"] = 0,
        }, -- [4]
        {
            ["incrementalCritical"] = 0,
            ["totalTime"] = 2.634000062942505,
            ["masteryTotal"] = 0,
            ["incrementalMastery"] = 0,
            ["effectiveTotal"] = 8320,
            ["encounterName"] = "Ancient Mogu Spirit",
            ["incrementalCriticalAll"] = 0,
            ["isBoss"] = false,
            ["startTime"] = 0,
            ["weightedTotal"] = 131.1201653246354,
            ["criticalTotal"] = 0,
        }, -- [5]
        {
            ["incrementalCritical"] = 0,
            ["totalTime"] = 2.644999980926514,
            ["masteryTotal"] = 676.9085283958921,
            ["incrementalMastery"] = 0.0498313123266094,
            ["effectiveTotal"] = 39147,
            ["encounterName"] = "Ancient Mogu Spirit",
            ["isBoss"] = false,
            ["weightedTotal"] = 1184.969985706814,
            ["startTime"] = 0,
            ["incrementalCriticalAll"] = 0.04386666666666667,
            ["criticalTotal"] = 0,
        }, -- [6]
        ["version"] = 18,
    }

    MigrateArchive(testArchive)

    if not testArchive then return false; end
    if not testArchive.version == archiveVersion then return false; end
    r = rss.CreateAggregateRecord("Test", testArchive, function(r) return true; end)
    if not r then return false; end
    if not (r.effectiveTotal > 0.) then return false; end;
    return true;
end
